/*
 * New Network code -- In progress
 *
 *  The In Progress part means it is far from fully functional, so do not
 *   expect too much of it.
 *
 * What works?
 *  - Basis operations
 *
 * Known bugs
 *  - None right now
 *
 * Todo
 *  - Send player seed with every command so all commands are done the same
 *      both server and clients
 *  - Fix the GUI so people can enter their playername when joining
 *
 */

#include "network_data.h"
#include "network_client.h"
#include "network_server.h"
#include "variables.h"
#include "gui.h" /* XXX ShowErrorMessage */
#include "console.h" /* XXX IConsoleCmdExec */
#include "functions.h"
#include <stdarg.h> /* va_list */


#ifdef ENABLE_NETWORK
// The listen socket for the server
static SOCKET _listensocket;

// The amount of clients connected
static byte _network_clients_connected = 0;
// The index counter for new clients (is never decreased)
static uint16 _network_client_index = 0;

void NetworkGetClientName(char *clientname, size_t size, ClientState *cs)
{
	NetworkClientInfo *ci = DEREF_CLIENT_INFO(cs);
	if (ci->client_name[0] == '\0')
		snprintf(clientname, size, "Client #%d", cs->index);
	else
		snprintf(clientname, size, "%s", ci->client_name);
}

void NetworkTextMessage(NetworkAction action, const char *name, const char *str, ...)
{
	char buf[1024];
	va_list va;

	va_start(va, str);
	vsprintf(buf, str, va);
	va_end(va);

	if (action == NETWORK_ACTION_JOIN_LEAVE)
		IConsolePrintF(10, "*** %s %s", name, buf);
	else
		IConsolePrintF(10, "%s: %s", name, buf);
}

// Calculate the frame-lag of a client
uint NetworkCalculateLag(ClientState *cs)
{
	int lag = cs->last_frame_server - cs->last_frame;
	// This client has missed his ACK packet after 1 DAY_TICKS..
	//  so we increase his lag for every frame that passes!
	// The packet can be out by a max of _net_frame_freq
	if (cs->last_frame_server + DAY_TICKS + _network_frame_freq < _frame_counter)
		lag += _frame_counter - (cs->last_frame_server + DAY_TICKS + _network_frame_freq);

	return lag;
}


// There was a non-recoverable error, drop back to the main menu with a nice
//  error
void NetworkError(StringID error_string)
{
	_switch_mode = SM_MENU;
	_switch_mode_errorstr = error_string;
}

void ClientStartError(char *error) {
		DEBUG(net, 0)("%s",error);
	_switch_mode = SM_MENU;
	_switch_mode_errorstr = STR_NETWORK_ERR_CLIENT_START;
}

void ServerStartError(char *error) {
	DEBUG(net, 0)("%s",error);
	_switch_mode = SM_MENU;
	_switch_mode_errorstr = STR_NETWORK_ERR_SERVER_START;
}

void NetworkClientError(byte res, ClientState *cs) {
	// First, send a CLIENT_ERROR to the server, so he knows we are
	//  disconnection (and why!)
	byte errorno;
	switch(res) {
		case NETWORK_RECV_STATUS_DESYNC: errorno = NETWORK_ERROR_DESYNC; break;
		case NETWORK_RECV_STATUS_SAVEGAME: errorno = NETWORK_ERROR_SAVEGAME_FAILED; break;
		default: errorno = NETWORK_ERROR_GENERAL;
	}
	// This means we fucked up and the server closed the connection
	if (res != NETWORK_RECV_STATUS_SERVER_ERROR) {
		SEND_COMMAND(PACKET_CLIENT_ERROR)(errorno);

		// Dequeue all commands before closing the socket
		NetworkSend_Packets(DEREF_CLIENT(0));
	}

	_switch_mode = SM_MENU;
	CloseClient(cs);
	_networking = false;
}

#ifdef _XBOX
static char ntoabuf[20];
char* inet_ntoa(struct   in_addr in) {
	sprintf(ntoabuf,"%d.%d.%d.%d",in.S_un.S_un_b.s_b1,in.S_un.S_un_b.s_b2,in.S_un.S_un_b.s_b3,in.S_un.S_un_b.s_b4);
	return ntoabuf;
}

int gethostname(char* name,int namelen) {
	XNADDR pxna;
	int retval=0;
	retval=XNetGetTitleXnAddr(&pxna);
	if(!(retval&XNET_GET_XNADDR_DHCP)&&!(retval&XNET_GET_XNADDR_STATIC)&&!(retval&XNET_GET_XNADDR_PPPOE))
		return -1;
	snprintf(name,namelen,"%d.%d.%d.%d",pxna.ina.S_un.S_un_b.s_b1,pxna.ina.S_un.S_un_b.s_b2,pxna.ina.S_un.S_un_b.s_b3,pxna.ina.S_un.S_un_b.s_b4);
	return 0;
}

XNDNS* XResolveHost(const char *host) {
	// hostname not found in cache.
	// do a DNS lookup
	WSAEVENT hEvent = WSACreateEvent();
	XNDNS* pDns		  = NULL;
	INT err = XNetDnsLookup(host, hEvent, &pDns);
	WaitForSingleObject( (HANDLE)hEvent, INFINITE);
	if( pDns && pDns->iStatus == 0 )
	{
		//DNS lookup succeeded
//		strIpAdres.Format("%d.%d.%d.%d",(ulHostIp & 0xFF),(ulHostIp & 0xFF00) >> 8,(ulHostIp & 0xFF0000) >> 16,(ulHostIp & 0xFF000000) >> 24 );
//		XNetDnsRelease(pDns);
		WSACloseEvent(hEvent);
		return pDns;
	}
	if (pDns)
	{
//		CLog::Log(LOGERROR, "DNS lookup for %s failed: %u", strHostName.c_str(), pDns->iStatus);
		XNetDnsRelease(pDns);
	}
//	else
//		CLog::Log(LOGERROR, "DNS lookup for %s failed: %u", strHostName.c_str(), err);
	WSACloseEvent(hEvent);
	return 0;
}

#endif


// Resolve a hostname to a inet_addr
unsigned long NetworkResolveHost(const char *hostname)
{
	in_addr_t ip;

	// First try: is it an ip address?
	ip = inet_addr(hostname);

	// If not try to resolve the name
	if (ip == INADDR_NONE) {
#ifdef _XBOX
		XNDNS *pxna;
		pxna=XResolveHost(hostname);
		if(!pxna) {
			DEBUG(net, 1) ("[NET][IP] cannot resolve %s", hostname);
		}
		else {
			ip=pxna->aina[0].S_un.S_addr;
			XNetDnsRelease(pxna);
		}
#else
		struct hostent *he = gethostbyname(hostname);
		if (he == NULL) {
			DEBUG(net, 0) ("[NET] Cannot resolve %s", hostname);
		} else {
			struct in_addr addr = *(struct in_addr *)he->h_addr_list[0];
			DEBUG(net, 1) ("[NET] Resolved %s to %s", hostname, inet_ntoa(addr));
			ip = addr.s_addr;
		}
#endif
	}
	return ip;
}

// Converts a string to ip/port/player
//  Format: IP#player:port
//
// connection_string will be re-terminated to seperate out the hostname, and player and port will
// be set to the player and port strings given by the user, inside the memory area originally
// occupied by connection_string.
void ParseConnectionString(const byte **player, const byte **port, byte *connection_string)
{
	byte *p;
	for (p = connection_string; *p != '\0'; p++) {
		if (*p == '#') {
			*player = p + 1;
			*p = '\0';
		} else if (*p == ':') {
			*port = p + 1;
			*p = '\0';
		}
	}
}

// Creates a new client from a socket
//   Used both by the server and the client
static ClientState *AllocClient(SOCKET s)
{
	ClientState *cs;
	NetworkClientInfo *ci;
	byte client_no;

	client_no = 0;

	if (_network_server) {
		// Can we handle a new client?
		if (_network_clients_connected >= MAX_CLIENTS)
			return NULL;

		if (_network_game_info.clients_on >= _network_game_info.clients_max)
			return NULL;

		// Register the login
		client_no = _network_clients_connected++;
		_network_game_info.clients_on++;
	}

	cs = &_clients[client_no];
	ci = &_network_client_info[client_no];
	memset(cs, 0, sizeof(*cs));
	memset(ci, 0, sizeof(*ci));
	cs->socket = s;
	cs->index = _network_client_index++;
	cs->last_frame = 0;
	cs->quited = false;
	return cs;
}

// Close a connection
void CloseClient(ClientState *cs)
{
	NetworkClientInfo *ci;
	// Socket is already dead
	if (cs->socket == INVALID_SOCKET) return;

	DEBUG(net, 1) ("[NET] Closed client connection");

	if (!cs->quited && _network_server) {
		// We did not receive a leave message from this client...
		byte errorno = NETWORK_ERROR_CONNECTION_LOST;
		char str1[100], str2[100];
		char clientname[100];
		ClientState *new_cs;

		NetworkGetClientName(clientname, 100, cs);

		GetString(str1, STR_NETWORK_ERR_LEFT);
		GetString(str2, STR_NETWORK_ERR_CLIENT_GENERAL + errorno);

		NetworkTextMessage(NETWORK_ACTION_JOIN_LEAVE, clientname, "%s (%s)", str1, str2);

		// Inform other clients of this... strange leaving ;)
		FOR_ALL_CLIENTS(new_cs) {
			if (new_cs->status > STATUS_AUTH) {
				SEND_COMMAND(PACKET_SERVER_ERROR_QUIT)(new_cs, clientname, errorno);
			}
		}
	}

	closesocket(cs->socket);

	// Free all pending and partially received packets
	while (cs->packet_queue != NULL) {
		Packet *p = cs->packet_queue->next;
		free(cs->packet_queue);
		cs->packet_queue = p;
	}
	free(cs->packet_recv);
	cs->packet_recv = NULL;

	while (cs->command_queue != NULL) {
		CommandPacket *p = cs->command_queue->next;
		free(cs->command_queue);
		cs->command_queue = p;
	}

	// Close the gap in the client-list
	ci = &_network_client_info[cs - _clients];
	while ((cs + 1) != DEREF_CLIENT(MAX_CLIENTS) && (cs + 1)->socket != INVALID_SOCKET) {
		*cs = *(cs + 1);
		*ci = *(ci + 1);
		cs++;
		ci++;
	}
	cs->socket = INVALID_SOCKET;
	cs->status = STATUS_INACTIVE;

	if (_network_server) {
		_network_game_info.clients_on--;
		_network_clients_connected--;
	}
}

// For the server, to accept new clients
static void NetworkAcceptClients()
{
	struct sockaddr_in sin;
	SOCKET s;
	ClientState *cs;
#ifndef __MORPHOS__
	int sin_len;
#else
	LONG sin_len; // for some reason we need a 'LONG' under MorphOS
#endif

	// Should never ever happen.. is it possible??
	assert(_listensocket != INVALID_SOCKET);

	for(;;) {
		sin_len = sizeof(sin);
		s = accept(_listensocket, (struct sockaddr*)&sin, &sin_len);
		if (s == INVALID_SOCKET) return;

		// set nonblocking mode for client socket
		{ unsigned long blocking = 1; ioctlsocket(s, FIONBIO, &blocking); }

		DEBUG(net, 1) ("[NET] Got client from %s on frame %d", inet_ntoa(sin.sin_addr), _frame_counter);

		// set nodelay
		// The (const char*) cast is needed for windows!!
		{int b = 1; setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (const char*)&b, sizeof(b));}

		cs = AllocClient(s);
		if (cs == NULL) {
			// no more clients allowed?
			closesocket(s);
			continue;
		}

		// a new client has connected. We set him at inactive for now
		//  maybe he is only requesting server-info. Till he has sent a PACKET_CLIENT_MAP_OK
		//  the client stays inactive
		cs->status = STATUS_INACTIVE;
		{
			NetworkClientInfo *ci;
			ci = DEREF_CLIENT_INFO(cs);
			ci->join_date = _date;
			snprintf(ci->client_ip, 15, "%s", inet_ntoa(sin.sin_addr));
		}
	}
}

// A client wants to connect to a server
bool NetworkConnect(const char *hostname, int port)
{
	SOCKET s;
	struct sockaddr_in sin;

	DEBUG(net, 1) ("[NET] Connecting to %s %d", hostname, port);

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == INVALID_SOCKET) {
		ClientStartError("socket() failed");
		return false;
	}

	{ // set nodelay /* XXX should this be done at all? */
		int b = 1;
		// The (const char*) cast is needed for windows!!
		if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (const char*)&b, sizeof(b)) != 0)
			DEBUG(net, 1)("[NET] Setting TCP_NODELAY failed");
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = NetworkResolveHost(hostname);
	sin.sin_port = htons(port);

	if (connect(s, (struct sockaddr*) &sin, sizeof(sin)) != 0) {
		// We failed to connect for which reason what so ever
		return false;
	}

	{ // set nonblocking mode for socket..
		unsigned long blocking = 1;
		if (ioctlsocket(s, FIONBIO, &blocking) != 0)
			DEBUG(net, 0)("[NET] Setting non-blocking failed"); /* XXX should this be an error? */
	}

	// in client mode, only the first client field is used. it's pointing to the server.
	AllocClient(s);

	return true;
}

// Set up the listen socket for the server
void NetworkListen(void)
{
	SOCKET ls;
	struct sockaddr_in sin;
	int port;

	port = _network_server_port;

	DEBUG(net, 1) ("[NET] Listening on port %d", port);

	ls = socket(AF_INET, SOCK_STREAM, 0);
	if (ls == INVALID_SOCKET) {
		// FIXME: Should this be a nice error?
		ServerStartError("socket() on listen socket failed");
		return;
	}

	{ // reuse the socket
		int reuse = 1;
		// The (const char*) cast is needed for windows!!
		if (setsockopt(ls, SOL_SOCKET, SO_REUSEADDR, (const char*)&reuse, sizeof(reuse)) == -1) {
			ServerStartError("setsockopt() on listen socket failed");
			return;
		}
	}

	{ // set nonblocking mode for socket
		unsigned long blocking = 1;
		if (ioctlsocket(ls, FIONBIO, &blocking) != 0)
			DEBUG(net, 0)("[NET] Setting non-blocking failed"); /* XXX should this be an error? */
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = 0;
	sin.sin_port = htons(port);

	if (bind(ls, (struct sockaddr*)&sin, sizeof(sin)) != 0) {
		ServerStartError("bind() failed");
		return;
	}

	if (listen(ls, 1) != 0) {
		ServerStartError("listen() failed");
		return;
	}

	_listensocket = ls;
}

// Close all current connections
void NetworkClose(void)
{
	ClientState *cs;

	FOR_ALL_CLIENTS(cs) {
		if (_network_server) {
		} else {
			SEND_COMMAND(PACKET_CLIENT_QUIT)("leaving");
			NetworkSend_Packets(cs);
		}
		CloseClient(cs);
	}

	if (_network_server) {
		// We are a server, also close the listensocket
		closesocket(_listensocket);
		_listensocket = INVALID_SOCKET;
		DEBUG(net, 1) ("[NET] Closed listener");
	}
}

// Inits the network (cleans sockets and stuff)
void NetworkInitialize(void)
{
	ClientState *cs;

	_network_game_list = NULL;
	_local_command_queue = NULL;

	// Clean all client-sockets
	FOR_ALL_CLIENTS(cs) {
		cs->socket = INVALID_SOCKET;
		cs->status = STATUS_INACTIVE;
		cs->command_queue = NULL;
	}

	_sync_frame = 0;
}

// Used by clients, to connect to a server
bool NetworkClientConnectGame(const byte* host, unsigned short port)
{
	if (!_network_available) return false;

	if (port == 0) {
		// Autodetect failed (if UDP was in place ;))
		ShowErrorMessage(-1, STR_NETWORK_ERR_NOSERVER, 0, 0);
		_switch_mode_errorstr = STR_NETWORK_ERR_NOSERVER;
		return false;
	}

	NetworkInitialize();

	_network_server = false;

	// Try to connect
	_networking = NetworkConnect(host, port);

	// We are connected
	if (_networking) {
		IConsoleCmdExec("exec scripts/on_client.scr 0");
		NetworkClient_Connected();
	} else {
		// Connecting failed
		ShowErrorMessage(-1, STR_NETWORK_ERR_NOCONNECTION,0,0);
		_switch_mode_errorstr = STR_NETWORK_ERR_NOCONNECTION;
	}

	return _networking;
}


bool NetworkServerStart(void)
{
	if (!_network_available) return false;

	NetworkInitialize();
	NetworkListen();

	_network_server = true;
	_networking = true;
	_frame_counter = 0;
	_frame_counter_server = 0;
	_frame_counter_max = 0;

	// The server is a client too ;)
	// We should make a nice function which fills all those things
	//  for now, this will do nicely ;)
	_network_game_info.clients_on = 1;
	_network_game_info.clients_max = MAX_CLIENTS;

	// execute server initialization script
	IConsoleCmdExec("exec scripts/on_server.scr 0");

	return true;
}

// We want to disconnect from the host/clients
void NetworkDisconnect(void)
{
	NetworkClose();

	// Free all queued commands
	while (_local_command_queue != NULL) {
		CommandPacket *p = _local_command_queue;
		_local_command_queue = _local_command_queue->next;
		free(p);
	}

	_networking = false;
	_network_server = false;
}

// Receives something from the network
bool NetworkReceive(void)
{
	ClientState *cs;
	int n;
	fd_set read_fd, write_fd;
	struct timeval tv;

	FD_ZERO(&read_fd);
	FD_ZERO(&write_fd);

	FOR_ALL_CLIENTS(cs) {
		FD_SET(cs->socket, &read_fd);
		FD_SET(cs->socket, &write_fd);
	}

	// take care of listener port
	if (_network_server) {
		FD_SET(_listensocket, &read_fd);
	}

	tv.tv_sec = tv.tv_usec = 0; // don't block at all.
#if !defined(__MORPHOS__) && !defined(__AMIGA__)
	n = select(FD_SETSIZE, &read_fd, &write_fd, NULL, &tv);
#else
	n = WaitSelect(FD_SETSIZE, &read_fd, &write_fd, NULL, &tv, NULL);
#endif
	if (n == -1 && !_network_server) NetworkError(STR_NETWORK_ERR_LOSTCONNECTION);

	// accept clients..
	if (_network_server && FD_ISSET(_listensocket, &read_fd))
		NetworkAcceptClients();

	// read stuff from clients
	FOR_ALL_CLIENTS(cs) {
		cs->writable = !!FD_ISSET(cs->socket, &write_fd);
		if (FD_ISSET(cs->socket, &read_fd)) {
			if (_network_server)
				NetworkServer_ReadPackets(cs);
			else {
				byte res;
				if ((res = NetworkClient_ReadPackets(cs)) != NETWORK_RECV_STATUS_OKAY) {
					// The client made an error of which we can not recover
					//   close the client and drop back to main menu

					NetworkClientError(res, cs);
					return false;
				}
			}
		}
	}
	return true;
}

// This sends all buffered commands (if possible)
static void NetworkSend(void)
{
	ClientState *cs;
	FOR_ALL_CLIENTS(cs) {
		if (cs->writable) {
			NetworkSend_Packets(cs);

			if (cs->status == STATUS_MAP) {
				// This client is in the middle of a map-send, call the function for that
				SEND_COMMAND(PACKET_SERVER_MAP)(cs);
			}
		}
	}
}

void NetworkHandleLocalQueue() {
	if (_local_command_queue != NULL) {
		CommandPacket *cp;
		CommandPacket *cp_prev;

		cp = _local_command_queue;
		cp_prev = NULL;

		while (cp != NULL) {
			if (_frame_counter > cp->frame) {
				NetworkExecuteCommand(cp);

				if (cp_prev != NULL) {
					cp_prev->next = cp->next;
					free(cp);
					cp = cp_prev->next;
				} else {
					// This means we are at our first packet
					_local_command_queue = cp->next;
					free(cp);
					cp = _local_command_queue;
				}

			} else {
				cp_prev = cp;
				cp = cp->next;
			}
		}
	}
}


extern void StateGameLoop();

bool NetworkDoClientLoop()
{
	_frame_counter++;

	NetworkHandleLocalQueue();

	StateGameLoop();

	// Check if we are in sync!
	if (_sync_frame != 0) {
		if (_sync_frame == _frame_counter) {
#ifdef NETWORK_SEND_DOUBLE_SEED
			if (_sync_seed_1 != _random_seeds[0][0] || _sync_seed_2 != _random_seeds[0][1]) {
#else
			if (_sync_seed_1 != _random_seeds[0][0]) {
#endif
				_switch_mode_errorstr = STR_NETWORK_ERR_DESYNC;
				DEBUG(net, 0)("[NET] Sync error detected!");
				NetworkClientError(NETWORK_RECV_STATUS_DESYNC, DEREF_CLIENT(0));
				return false;
			}
			_sync_frame = 0;
		} else if (_sync_frame < _frame_counter) {
			DEBUG(net, 1)("[NET] Missed frame for sync-test (%d / %d)", _sync_frame, _frame_counter);
			_sync_frame = 0;
		}
	}

	return true;
}

// The main loop called from ttd.c
//  Here we also have to do StateGameLoop if needed!
void NetworkGameLoop(void)
{
	if (!_networking) return;

	if (!NetworkReceive()) return;

	if (_network_server) {
		// We first increase the _frame_counter
		_frame_counter++;

		NetworkHandleLocalQueue();

		// Then we make the frame
		StateGameLoop();

		_sync_seed_1 = _random_seeds[0][0];
#ifdef NETWORK_SEND_DOUBLE_SEED
		_sync_seed_2 = _random_seeds[0][1];
#endif

		NetworkServer_Tick();
	} else {
		// Client

		// Make sure we are at the frame were the server is (quick-frames)
		if (_frame_counter_server > _frame_counter) {
			while (_frame_counter_server > _frame_counter) {
				if (!NetworkDoClientLoop()) break;
			}
		} else {
			// Else, keep on going till _frame_counter_max
			if (_frame_counter_max > _frame_counter) {
				NetworkDoClientLoop();
			}
		}
	}

	NetworkSend();
}






// This tries to launch the network for a given OS
void NetworkStartUp()
{
	DEBUG(net, 3) ("[NET][Core] Starting network...");
	// Network is available
	_network_available = true;
	_network_dedicated = false;

	// Let's load the network in windows
	#if defined(WIN32)
	{
		WSADATA wsa;
#ifdef _XBOX
		XNetStartupParams xnsp;
		int err;
		DEBUG(net, 3) ("[NET][Core] Using xbox network library");
		memset(&xnsp, 0, sizeof(xnsp));
		xnsp.cfgSizeOfStruct = sizeof(XNetStartupParams);
		xnsp.cfgFlags = XNET_STARTUP_BYPASS_SECURITY;
		err = XNetStartup(&xnsp);
		if(err) {
			DEBUG(net, 0) ("[NET][Core] Error: XNetStartup failed. Network not available.");
			_network_available = false;
			return;
		}
#else
		DEBUG(net, 3) ("[NET][Core] Loading windows socket library");
#endif
		if (WSAStartup(MAKEWORD(2,0), &wsa) != 0) {
			DEBUG(net, 0) ("[NET][Core] Error: WSAStartup failed. Network not available.");
			_network_available = false;
			return;
		}
	}
	#else
		#if defined(__MORPHOS__) || defined(__AMIGA__)
		{
			DEBUG(misc,3) ("[NET][Core] Loading bsd socket library");
			if (!(SocketBase = OpenLibrary("bsdsocket.library", 4))) {
				DEBUG(net, 0) ("[NET][Core] Error: couldn't open bsdsocket.library version 4. Network not available.");
				_network_available = false;
				return;
			}

			#if defined(__AMIGA__)
			// for usleep() implementation (only required for legacy AmigaOS builds)
			if ( (TimerPort = CreateMsgPort()) ) {
				if ( (TimerRequest = (struct timerequest *) CreateIORequest(TimerPort, sizeof(struct timerequest))) ) {
					if ( OpenDevice("timer.device", UNIT_MICROHZ, (struct IORequest *) TimerRequest, 0) == 0 ) {
						if ( !(TimerBase = TimerRequest->tr_node.io_Device) ) {
							// free ressources...
							DEBUG(net, 0) ("[NET][Core] Error: couldn't initialize timer. Network not available.");
							_network_available = false;
							return;
						}
					}
				}
			}
			#endif // __AMIGA__
		}
		#endif // __MORPHOS__ / __AMIGA__
	#endif

	NetworkInitialize();
	DEBUG(net, 3) ("[NET][Core] Network online. Multiplayer available.");
}

// This shuts the network down
void NetworkShutDown(void)
{
	DEBUG(net, 3) ("[NET][Core] Shutting down the network.");

	_network_available = false;

	#if defined(__MORPHOS__) || defined(__AMIGA__)
	{
		// free allocated ressources
		#if !defined(__MORPHOS__)
			if (TimerBase)    { CloseDevice((struct IORequest *) TimerRequest); }
			if (TimerRequest) { DeleteIORequest(TimerRequest); }
			if (TimerPort)    { DeleteMsgPort(TimerPort); }
		#endif

		if (SocketBase) {
			CloseLibrary(SocketBase);
		}
	}
	#endif

	#if defined(WIN32)
	{
		WSACleanup();
	}
	#endif
}


#else // ENABLE_NETWORK
#endif
