#include "network_server.h"
#include "variables.h" /* XXX _frame_counter */
#include "console.h"

// This file handles all the server-commands

// Is the network enabled?
#ifdef ENABLE_NETWORK


// **********
// Sending functions
//   DEF_SERVER_SEND_COMMAND has parameter: ClientState *cs
// **********

DEF_SERVER_SEND_COMMAND(PACKET_SERVER_GAME_INFO)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_GAME_INFO);
	// TODO: send the GAME_INFO

	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_ERROR)(ClientState *cs, NetworkErrorCode error)
{
	ClientState *new_cs;
	char str1[100], str2[100];
	char clientname[100];

	Packet *p = NetworkSend_Init(PACKET_SERVER_ERROR);
	NetworkSend_uint8(p, error);
	NetworkSend_Packet(p, cs);

	NetworkGetClientName(clientname, 100, cs);

	GetString(str1, STR_NETWORK_ERR_LEFT);
	GetString(str2, STR_NETWORK_ERR_CLIENT_GENERAL + error);

	DEBUG(net, 2)("[NET] %s made an error (%s) and his connection is closed", clientname, str2);

	NetworkTextMessage(NETWORK_ACTION_JOIN_LEAVE, clientname, "%s (%s)", str1, str2);

	// Only send when the current client was in game
	if (cs->status > STATUS_AUTH) {
		FOR_ALL_CLIENTS(new_cs) {
			if (new_cs->status > STATUS_AUTH && new_cs != cs) {
				// Some errors we filter to a more general error. Clients don't have to know the real
				//  reason a joining failed.
				if (error == NETWORK_ERROR_NOT_AUTHORIZED || error == NETWORK_ERROR_NOT_EXPECTED || error == NETWORK_ERROR_WRONG_REVISION)
					error = NETWORK_ERROR_ILLEGAL_PACKET;

				SEND_COMMAND(PACKET_SERVER_ERROR_QUIT)(new_cs, clientname, error);
			}
		}
	}

	cs->quited = true;

	// Make sure the data get's there before we close the connection
	NetworkSend_Packets(cs);

	// The client made a mistake, so drop his connection now!
	CloseClient(cs);
}

DEF_SERVER_SEND_COMMAND(PACKET_SERVER_WELCOME)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_WELCOME);
	NetworkSend_Packet(p, cs);
}

// This sends the map to the client
DEF_SERVER_SEND_COMMAND(PACKET_SERVER_MAP)
{
	char filename[256];
	static FILE *file_pointer;
	static uint sent_packets; // How many packets we did send succecfully last time

	if (cs->status < STATUS_AUTH) {
		// Illegal call, return error and ignore the packet
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_NOT_AUTHORIZED);
		return;
	}
	if (cs->status == STATUS_AUTH) {
		Packet *p;
		// Make a dump of the current game
		sprintf(filename, "%s%snetwork_server.tmp",  _path.autosave_dir, PATHSEP);
		if (SaveOrLoad(filename, SL_SAVE) != SL_OK) error("network savedump failed");

		file_pointer = fopen(filename, "rb");

		// Now send the _frame_counter and how many packets are coming
		p = NetworkSend_Init(PACKET_SERVER_MAP);
		NetworkSend_uint8(p, MAP_PACKET_START);
		NetworkSend_uint32(p, _frame_counter);
		NetworkSend_Packet(p, cs);

		sent_packets = 4; // We start with trying 4 packets

		cs->status = STATUS_MAP;
	}

	if (cs->status == STATUS_MAP) {
		uint i;
		int res;
		for (i=0;i<sent_packets;i++) {
			Packet *p = NetworkSend_Init(PACKET_SERVER_MAP);
			NetworkSend_uint8(p, MAP_PACKET_NORMAL);
			res = fread(p->buffer + p->size, 1, SEND_MTU - p->size, file_pointer);
			if (ferror(file_pointer)) {
				error("Error reading temporary network savegame!");
			}
			p->size += res;
			NetworkSend_Packet(p, cs);
			if (feof(file_pointer)) {
				// Done reading!
				// TODO: dequeue buffered commands sent while client was receiving map
				int i;
				Packet *p = NetworkSend_Init(PACKET_SERVER_MAP);
				NetworkSend_uint8(p, MAP_PACKET_END);
				// Send the player_seeds in this packet
				for (i=0;i<MAX_PLAYERS;i++) {
					NetworkSend_uint32(p, _player_seeds[i][0]);
					NetworkSend_uint32(p, _player_seeds[i][1]);
				}
				NetworkSend_Packet(p, cs);

				// Set the status to DONE_MAP, no we will wait for the client
				//  to send it is ready (maybe that happens like never ;))
				cs->status = STATUS_DONE_MAP;
				fclose(file_pointer);

				// There is no more data, so break the for
				break;
			}
		}

		// Send all packets (forced) and check if we have send it all
		NetworkSend_Packets(cs);
		if (cs->packet_queue == NULL) {
			// All are sent, increase the sent_packets
			sent_packets *= 2;
		} else {
			// Not everything is sent, decrease the sent_packets
			if (sent_packets > 1) sent_packets /= 2;
		}
	}
}


// This very simple sends the _frame_counter
DEF_SERVER_SEND_COMMAND(PACKET_SERVER_FRAME)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_FRAME);
	NetworkSend_uint32(p, _frame_counter);
	NetworkSend_uint32(p, _frame_counter_max);
#ifdef ENABLE_NETWORK_SYNC_EVERY_FRAME
	NetworkSend_uint32(p, _sync_seed_1);
#ifdef NETWORK_SEND_DOUBLE_SEED
	NetworkSend_uint32(p, _sync_seed_2);
#endif
#endif
	NetworkSend_Packet(p, cs);
}

// This function sends our seed to the client, to check if it is still in sync
//  the frame_counter is needed to check if the client is on the right frame
DEF_SERVER_SEND_COMMAND(PACKET_SERVER_SYNC)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_SYNC);
	NetworkSend_uint32(p, _frame_counter);
	NetworkSend_uint32(p, _sync_seed_1);
#ifdef NETWORK_SEND_DOUBLE_SEED
	NetworkSend_uint32(p, _sync_seed_2);
#endif
	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_COMMAND)(ClientState *cs, CommandPacket *cp)
{
	int i;
	Packet *p = NetworkSend_Init(PACKET_SERVER_COMMAND);

	NetworkSend_uint8(p, cp->player);
	NetworkSend_uint32(p, cp->cmd);
	NetworkSend_uint32(p, cp->p1);
	NetworkSend_uint32(p, cp->p2);
	NetworkSend_uint32(p, (uint32)cp->tile);
	for (i=0;i<lengthof(cp->dp);i++) {
		NetworkSend_uint32(p, cp->dp[i]);
	}
	NetworkSend_uint8(p, cp->callback);
	NetworkSend_uint32(p, cp->frame);

	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_ERROR_QUIT)(ClientState *cs, const char *clientname, byte errorno)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_ERROR_QUIT);

	NetworkSend_string(p, clientname);
	NetworkSend_uint8(p, errorno);

	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_QUIT)(ClientState *cs, const char *clientname, const char *leavemsg)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_QUIT);

	NetworkSend_string(p, clientname);
	NetworkSend_string(p, leavemsg);

	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_JOIN)(ClientState *cs, const char *clientname)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_JOIN);

	NetworkSend_string(p, clientname);

	NetworkSend_Packet(p, cs);
}

DEF_SERVER_SEND_COMMAND_PARAM(PACKET_SERVER_CHAT)(ClientState *cs, NetworkAction action, DestType desttype, const char *name, const char *msg)
{
	Packet *p = NetworkSend_Init(PACKET_SERVER_CHAT);

	NetworkSend_uint8(p, action);
	NetworkSend_uint8(p, desttype);
	NetworkSend_string(p, name);
	NetworkSend_string(p, msg);

	NetworkSend_Packet(p, cs);
}

// **********
// Receiving functions
//   DEF_SERVER_RECEIVE_COMMAND has parameter: ClientState *cs, Packet *p
// **********

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_GAME_INFO)
{
#if defined(WITH_REV)
		extern char _openttd_revision[];
#else
		const char _openttd_revision[] = "norev000";
#endif
	// TODO: fix this to work with ServerInfo
	char client_revision[20];
	NetworkRecv_string(p, client_revision, 20);

	//  Too bad, when WITH_REV is disabled, we can not compare the version.
#if defined(WITH_REV)
	// Check if the client has WITH_REV enabled
	if (strncmp("norev000", client_revision, 20) != 0) {
		if (strncmp(_openttd_revision, client_revision, 20) != 0) {
			// Different revisions!!
			SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_WRONG_REVISION);

			return;
		}
	}
#endif

	SEND_COMMAND(PACKET_SERVER_GAME_INFO)(cs);
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_JOIN)
{
	char name[40];
	char password[100];
	NetworkClientInfo *ci;
	ClientState *new_cs;
	bool found_name = false;
	char test_name[45];
	byte number = 0;

	NetworkRecv_string(p, name, 40);
	NetworkRecv_string(p, password, 100);

	// Check game-password
	if (strncmp(password, _network_game_info.server_password, 20) != 0) {
		// Password is invalid
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_WRONG_PASSWORD);
		return;
	}

	// Check if someone else already has that name
	snprintf(test_name, 45, "%s", name);
	while (!found_name) {
		FOR_ALL_CLIENTS(new_cs) {
			ci = DEREF_CLIENT_INFO(new_cs);
			if (strncmp(ci->client_name, test_name, 40) == 0) {
				// Name already in use
				// Try a new name (<name> #1, <name> #2, and so on)

				// Stop if we tried for more then 50 times..
				if (number++ > 50) break;
				snprintf(test_name, 45, "%s #%d", name, number);
				break;
			} else {
				found_name = true;
			}
		}
		if (number > 50) break;
	}
	if (!found_name) {
		// We could not create a name for this player
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_NAME_IN_USE);
		return;
	}

	ci = DEREF_CLIENT_INFO(cs);

	snprintf(ci->client_name, 40, "%s", test_name);
	cs->status = STATUS_AUTH;

	SEND_COMMAND(PACKET_SERVER_WELCOME)(cs);
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_GETMAP)
{
	// The client was never joined.. so this is impossible, right?
	//  Ignore the packet, give the client a warning, and close his connection
	if (cs->status < STATUS_AUTH || cs->quited) {
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_NOT_AUTHORIZED);
		return;
	}

	// We receive a request to upload the map.. give it to the client!
	SEND_COMMAND(PACKET_SERVER_MAP)(cs);
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_MAP_OK)
{
	// Client has the map, now start syncing
	if (cs->status == STATUS_DONE_MAP && !cs->quited) {
		char clientname[100];
		char str[100];
		ClientState *new_cs;
		GetString(str, STR_NETWORK_CLIENT_JOINED);

		NetworkGetClientName(clientname, 100, cs);

		NetworkTextMessage(NETWORK_ACTION_JOIN_LEAVE, clientname, str);

		cs->status = STATUS_ACTIVE;

		// This is the frame the client receives
		//  we need it later on to make sure the client is not too slow
		cs->last_frame = _frame_counter;
		cs->last_frame_server = _frame_counter;

		FOR_ALL_CLIENTS(new_cs) {
			if (new_cs->status > STATUS_AUTH) {
				SEND_COMMAND(PACKET_SERVER_JOIN)(new_cs, clientname);
			}
		}
	} else {
		// Wrong status for this packet, give a warning to client, and close connection
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_NOT_EXPECTED);
	}
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_COMMAND)
{
	// The client has done a command and wants us to handle it
	int i;
	byte callback;
	ClientState *new_cs;

	CommandPacket *cp = malloc(sizeof(CommandPacket));

	// The client was never joined.. so this is impossible, right?
	//  Ignore the packet, give the client a warning, and close his connection
	if (cs->status < STATUS_DONE_MAP || cs->quited) {
		SEND_COMMAND(PACKET_SERVER_ERROR)(cs, NETWORK_ERROR_NOT_EXPECTED);
		return;
	}

	cp->player = NetworkRecv_uint8(p);
	cp->cmd = NetworkRecv_uint32(p);
	cp->p1 = NetworkRecv_uint32(p);
	cp->p2 = NetworkRecv_uint32(p);
	cp->tile = NetworkRecv_uint32(p);
	for (i=0;i<lengthof(cp->dp);i++)
		cp->dp[i] = NetworkRecv_uint32(p);

	callback = NetworkRecv_uint8(p);

	// The frame can be executed in the same frame as the next frame-packet
	//  That frame just before that frame is saved in _frame_counter_max
	cp->frame = _frame_counter_max + 1;
	cp->next = NULL;

	// Queue the command for the clients (are send at the end of the frame
	//   if they can handle it ;))
	FOR_ALL_CLIENTS(new_cs) {
		if (new_cs->status > STATUS_AUTH) {
			// Callbacks are only send back to the client who sent them in the
			//  first place. This filters that out.
			if (new_cs != cs)
				cp->callback = 0;
			else
				cp->callback = callback;
			NetworkAddCommandQueue(new_cs, cp);
		}
	}

	cp->callback = 0;
	// Queue the command on the server
	if (_local_command_queue == NULL) {
		_local_command_queue = cp;
	} else {
		// Find last packet
		CommandPacket *c = _local_command_queue;
		while (c->next != NULL) c = c->next;
		c->next = cp;
	}
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_ERROR)
{
	ClientState *new_cs;
	byte errorno = NetworkRecv_uint8(p);
	// A client made an error.. display the error
	char str1[100], str2[100];
	char clientname[100];

	// The client was never joined.. thank the client for the packet, but ignore it
	if (cs->status < STATUS_DONE_MAP || cs->quited) {
		cs->quited = true;
		return;
	}

	NetworkGetClientName(clientname, 100, cs);

	GetString(str1, STR_NETWORK_ERR_LEFT);
	GetString(str2, STR_NETWORK_ERR_CLIENT_GENERAL + errorno);

	DEBUG(net, 2)("[NET] %s reported an error and is closing his connection (%s)", clientname, str2);

	NetworkTextMessage(NETWORK_ACTION_JOIN_LEAVE, clientname, "%s (%s)", str1, str2);

	FOR_ALL_CLIENTS(new_cs) {
		if (new_cs->status > STATUS_AUTH) {
			SEND_COMMAND(PACKET_SERVER_ERROR_QUIT)(new_cs, clientname, errorno);
		}
	}

	cs->quited = true;
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_QUIT)
{
	ClientState *new_cs;
	// A client made an error.. display the error
	char str1[100], str2[100];
	char clientname[100];

	// The client was never joined.. thank the client for the packet, but ignore it
	if (cs->status < STATUS_DONE_MAP || cs->quited) {
		cs->quited = true;
		return;
	}

	NetworkRecv_string(p, str2, 100);

	NetworkGetClientName(clientname, 100, cs);

	GetString(str1, STR_NETWORK_ERR_LEFT);

	NetworkTextMessage(NETWORK_ACTION_JOIN_LEAVE, clientname, "%s (%s)", str1, str2);

	FOR_ALL_CLIENTS(new_cs) {
		if (new_cs->status > STATUS_AUTH) {
			SEND_COMMAND(PACKET_SERVER_QUIT)(new_cs, clientname, str2);
		}
	}

	cs->quited = true;
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_ACK)
{
	// The client received the frame, make note of it
	cs->last_frame = NetworkRecv_uint32(p);
	// With those 2 values we can calculate the lag realtime
	cs->last_frame_server = _frame_counter;
}



void NetworkServer_HandleChat(NetworkAction action, DestType desttype, int dest, const char *msg, const char *name)
{
	ClientState *cs;

	switch (desttype) {
#if 0 /* broken. fix later. */
	case DESTTYPE_CLIENT:
		NetworkSend_Packet (outp, DEREF_CLIENT (dest));
		break;
#endif
#if 0 /* broken. fix later. */
	case DESTTYPE_PLAYER:
		/* ... */
		break;
#endif
	default:
		DEBUG(net, 0)("[NET][Server] Received unknown destination type %d. Doing broadcast instead.\n");
		/* fall-through to next case */
	case DESTTYPE_BROADCAST:
		FOR_ALL_CLIENTS(cs) {
			SEND_COMMAND(PACKET_SERVER_CHAT)(cs, action, desttype, name, msg);
		}
		NetworkTextMessage(action, name, "%s", msg);
		break;
	}
}

DEF_SERVER_RECEIVE_COMMAND(PACKET_CLIENT_CHAT)
{
	NetworkAction action = NetworkRecv_uint8(p);
	DestType desttype = NetworkRecv_uint8(p);
	int dest = NetworkRecv_uint8(p);
	char msg[MAX_TEXT_MSG_LEN];
	char name[MAX_CLIENT_NAME_LEN];

	NetworkGetClientName(name, MAX_CLIENT_NAME_LEN, cs);

	NetworkRecv_string(p, msg, MAX_TEXT_MSG_LEN);

	NetworkServer_HandleChat(action, desttype, dest, msg, name);
}


// The layout for the receive-functions by the server
typedef void NetworkServerPacket(ClientState *cs, Packet *p);


// This array matches PacketType. At an incoming
//  packet it is matches against this array
//  and that way the right function to handle that
//  packet is found.
static NetworkServerPacket* const _network_server_packet[] = {
	RECEIVE_COMMAND(PACKET_CLIENT_GAME_INFO),
	NULL, /*PACKET_SERVER_GAME_INFO,*/
	RECEIVE_COMMAND(PACKET_CLIENT_JOIN),
	NULL, /*PACKET_SERVER_ERROR,*/
	NULL, /*PACKET_SERVER_WELCOME,*/
	RECEIVE_COMMAND(PACKET_CLIENT_GETMAP),
	NULL, /*PACKET_SERVER_MAP,*/
	RECEIVE_COMMAND(PACKET_CLIENT_MAP_OK),
	NULL, /*PACKET_SERVER_JOIN,*/
	NULL, /*PACKET_SERVER_FRAME,*/
	NULL, /*PACKET_SERVER_SYNC,*/
	RECEIVE_COMMAND(PACKET_CLIENT_ACK),
	RECEIVE_COMMAND(PACKET_CLIENT_COMMAND),
	NULL, /*PACKET_SERVER_COMMAND,*/
	RECEIVE_COMMAND(PACKET_CLIENT_CHAT),
	NULL, /*PACKET_SERVER_CHAT,*/
	RECEIVE_COMMAND(PACKET_CLIENT_QUIT),
	RECEIVE_COMMAND(PACKET_CLIENT_ERROR),
	NULL, /*PACKET_SERVER_QUIT,*/
	NULL, /*PACKET_SERVER_ERROR_QUIT,*/
	NULL, /*PACKET_SERVER_SHUTDOWN,*/
	NULL, /*PACKET_SERVER_NEWGAME,*/
};

/* Check if every packet type has a handler */
assert_compile(lengthof(_network_server_packet) == PACKET_END);

bool NetworkServer_ReadPackets(ClientState *cs)
{
	Packet *p;
	NetworkRecvStatus res;
	while((p = NetworkRecv_Packet(cs, &res)) != NULL) {
		byte type = NetworkRecv_uint8(p);
		if (type < PACKET_END && _network_server_packet[type] != NULL)
			_network_server_packet[type](cs, p);
		else
			DEBUG(net, 0)("[NET][Server] Received invalid packet type %d", type);
		free(p);
	}

	return true;
}

void NetworkHandleCommandQueue(ClientState *cs) {
	if (cs->command_queue != NULL) {
		CommandPacket *cp;
		CommandPacket *cp_prev;

		cp = cs->command_queue;
		cp_prev = NULL;

		while (cp != NULL) {
			if (_frame_counter > cp->frame) {
				SEND_COMMAND(PACKET_SERVER_COMMAND)(cs, cp);

				if (cp_prev != NULL) {
					cp_prev->next = cp->next;
					free(cp);
					cp = cp_prev->next;
				} else {
					// This means we are at our first packet
					cs->command_queue = cp->next;
					free(cp);
					cp = cs->command_queue;
				}

			} else {
				cp_prev = cp;
				cp = cp->next;
			}
		}
	}
}

void NetworkServer_Tick()
{
#ifndef ENABLE_NETWORK_SYNC_EVERY_FRAME
	static uint32 last_sync_frame = 0;
#endif
	ClientState *cs;
	bool send_frame = false;

	if (_frame_counter > _frame_counter_max) {
		_frame_counter_max = _frame_counter + _network_frame_freq;
		send_frame = true;
	}

	// Now we are done with the frame, inform the clients that they can
	//  do their frame!
	FOR_ALL_CLIENTS(cs) {
		// Check if the speed of the client is what we can expect from a client
		if (cs->status == STATUS_ACTIVE) {
			// 1 lag-point per day
			int lag = NetworkCalculateLag(cs) / DAY_TICKS;
			if (lag > 0) {
				if (lag > 3) {
					// Client did still not report in after 4 game-day, drop him
					//  (that is, the 3 of above, + 1 before any lag is counted)
					IConsolePrintF(_iconsole_color_error,"Client #%d is dropped because the client did not respond for more then 4 game-days", cs->index);
					CloseClient(cs);
					continue;
				}

				// Report once per time we detect the lag
				if (cs->lag_test == 0) {
					IConsolePrintF(_iconsole_color_warning,"[%d] Client #%d is slow, try increasing *net_frame_freq to a higher value!", _frame_counter, cs->index);
					cs->lag_test = 1;
				}
			} else {
				cs->lag_test = 0;
			}
		}


		// Check if we can send command, and if we have anything in the queue
		if (cs->status > STATUS_DONE_MAP) {
			NetworkHandleCommandQueue(cs);
		}

		// Do we need to send the new frame-packet?
		if (send_frame && cs->status == STATUS_ACTIVE) {
			SEND_COMMAND(PACKET_SERVER_FRAME)(cs);
		}
#ifndef ENABLE_NETWORK_SYNC_EVERY_FRAME
		// Is it time to send a sync-packet to all clients?
		if (last_sync_frame + _network_sync_freq < _frame_counter) {
			SEND_COMMAND(PACKET_SERVER_SYNC)(cs);
		}
#endif
	}

#ifndef ENABLE_NETWORK_SYNC_EVERY_FRAME
	// Update the last_sync_frame if needed!
	if (last_sync_frame + _network_sync_freq < _frame_counter) {
		last_sync_frame = _frame_counter;
	}
#endif
}

#endif // enable-network
