#include "stdafx.h"

#if defined(WITH_SDL)
#include "ttd.h"
#include "gfx.h"
#include "sound.h"
#include "window.h"
#include <SDL.h>
#include <SDL_mixer.h>
#include "player.h"
#include "hal.h"

#ifdef UNIX
#include <signal.h>
#endif

extern void mix_channels(void *udata, Uint8 *stream, int len);
static SDL_Joystick *joy[4];
static SDL_Surface *_sdl_screen;
static int _sdl_usage;
static bool _all_modes;

#define MAX_DIRTY_RECTS 100
static SDL_Rect _dirty_rects[MAX_DIRTY_RECTS];
static int _num_dirty_rects;

#define SDL_CALL

#if defined(DYNAMICALLY_LOADED_SDL) && defined(WIN32)


typedef struct {
	int (SDLCALL *SDL_Init)(Uint32);
	int (SDLCALL *SDL_InitSubSystem)(Uint32);
	char *(SDLCALL *SDL_GetError)();
	void (SDLCALL *SDL_QuitSubSystem)(Uint32);
	void (SDLCALL *SDL_UpdateRect)(SDL_Surface *, Sint32, Sint32, Uint32, Uint32);
	void (SDLCALL *SDL_UpdateRects)(SDL_Surface *, int, SDL_Rect *);
	int (SDLCALL *SDL_SetColors)(SDL_Surface *, SDL_Color *, int, int);
	void (SDLCALL *SDL_WM_SetCaption)(const char *, const char *);
	int (SDLCALL *SDL_ShowCursor)(int);
	void (SDLCALL *SDL_FreeSurface)(SDL_Surface *);
	int (SDLCALL *SDL_PollEvent)(SDL_Event *);
	void (SDLCALL *SDL_WarpMouse)(Uint16, Uint16);
	uint32 (SDLCALL *SDL_GetTicks)();
	int (SDLCALL *SDL_OpenAudio)(SDL_AudioSpec *, SDL_AudioSpec*);
	void (SDLCALL *SDL_PauseAudio)(int);
	void (SDLCALL *SDL_CloseAudio)();
	int (SDLCALL *SDL_LockSurface)(SDL_Surface*);
	void (SDLCALL *SDL_UnlockSurface)(SDL_Surface*);
	SDLMod (SDLCALL *SDL_GetModState)();
	void (SDLCALL *SDL_Delay)(Uint32);
	void (SDLCALL *SDL_Quit)();
	SDL_Surface *(SDLCALL *SDL_SetVideoMode)(int,int,int,Uint32);
	int (SDLCALL *SDL_EnableKeyRepeat)(int,int);
	void (SDLCALL *SDL_EnableUNICODE)(int);
	void (SDLCALL *SDL_VideoDriverName)(char *, int);
	SDL_Rect **(SDLCALL *SDL_ListModes)(void *, int);
	Uint8 *(SDLCALL *SDL_GetKeyState)(int *);
} SDLProcs;

#define M(x) x "\0"
static const char sdl_files[] =
	M("sdl.dll")
	M("SDL_Init")
	M("SDL_InitSubSystem")
	M("SDL_GetError")
	M("SDL_QuitSubSystem")
	M("SDL_UpdateRect")
	M("SDL_UpdateRects")
	M("SDL_SetColors")
	M("SDL_WM_SetCaption")
	M("SDL_ShowCursor")
	M("SDL_FreeSurface")
	M("SDL_PollEvent")
	M("SDL_WarpMouse")
	M("SDL_GetTicks")
	M("SDL_OpenAudio")
	M("SDL_PauseAudio")
	M("SDL_CloseAudio")
	M("SDL_LockSurface")
	M("SDL_UnlockSurface")
	M("SDL_GetModState")
	M("SDL_Delay")
	M("SDL_Quit")
	M("SDL_SetVideoMode")
	M("SDL_EnableKeyRepeat")
	M("SDL_EnableUNICODE")
	M("SDL_VideoDriverName")
	M("SDL_ListModes")
	M("SDL_GetKeyState")
	M("")
;
#undef M

static SDLProcs _proc;


#undef SDL_CALL
#define SDL_CALL _proc.

#endif


#ifdef UNIX
static void SdlAbort(int sig)
{
	/* Own hand-made parachute for the cases of failed assertions. */
	SDL_CALL SDL_Quit();
}
#endif


static char *SdlOpen(uint32 x)
{
#if defined(DYNAMICALLY_LOADED_SDL) && defined(WIN32)
	{ char *s = LoadSdlDLL(); if (s) return s; }
#endif
	if (_sdl_usage++ == 0) {
		if (SDL_CALL SDL_Init(x) == -1)
			return SDL_CALL SDL_GetError();
	} else if (x) {
		if (SDL_CALL SDL_InitSubSystem(x) == -1)
			return SDL_CALL SDL_GetError();
	}

#ifdef UNIX
	signal(SIGABRT, SdlAbort);
#endif

	return NULL;
}

static void SdlClose(uint32 x)
{
	if (x)
		SDL_CALL SDL_QuitSubSystem(x);
	if (--_sdl_usage == 0) {
		SDL_CALL SDL_Quit();
#ifdef UNIX
#ifndef __MORPHOS__
		signal(SIGABRT, SIG_DFL);
#else
		signal(SIGABRT, (void (*)(int))0 );
#endif
#endif
	}
}

static void SdlVideoMakeDirty(int left, int top, int width, int height)
{
//	printf("(%d,%d)-(%d,%d)\n", left, top, width, height);
//	_pixels_redrawn += width*height;
	if (_num_dirty_rects < MAX_DIRTY_RECTS) {
		_dirty_rects[_num_dirty_rects].x = left;
		_dirty_rects[_num_dirty_rects].y = top;
		_dirty_rects[_num_dirty_rects].w = width;
		_dirty_rects[_num_dirty_rects].h = height;
	}
	_num_dirty_rects++;
}

static SDL_Color pal[256];

static void UpdatePalette(uint start, uint end) {
	uint i;
	byte *b;

	for(i = start, b = _cur_palette + start * 3; i != end; i++, b += 3) {
		pal[i].r = b[0];
		pal[i].g = b[1];
		pal[i].b = b[2];
		pal[i].unused = b[3];
	}

	SDL_CALL SDL_SetColors(_sdl_screen, pal, start, end);
}

static void InitPalette(void) {
	UpdatePalette(0, 256);
}

static void CheckPaletteAnim()
{
	if(_pal_last_dirty != -1) {
		UpdatePalette(_pal_first_dirty, _pal_last_dirty + 1);
		_pal_last_dirty = -1;
	}
}

static void DrawSurfaceToScreen()
{
	int n;

	if ((n=_num_dirty_rects) != 0) {
		_num_dirty_rects = 0;
		if (n > MAX_DIRTY_RECTS)
			SDL_CALL SDL_UpdateRect(_sdl_screen, 0, 0, 0, 0);
		else {
			SDL_CALL SDL_UpdateRects(_sdl_screen, n, _dirty_rects);
		}
	}
}

static int CDECL compare_res(const void *pa, const void *pb)
{
	int x = ((const uint16*)pa)[0] - ((const uint16*)pb)[0];
	if (x) return x;
	return ((const uint16*)pa)[1] - ((const uint16*)pb)[1];
}

static const uint16 default_resolutions[][2] = {
	{640,480},
	{800,600},
	{1024,768},
	{1152,864},
	{1280,960},
	{1280,1024},
	{1400,1050},
	{1600,1200},
};

static void GetVideoModes(void) {
	int i;
	SDL_Rect **modes;

	modes = SDL_CALL SDL_ListModes(NULL, SDL_SWSURFACE + (_fullscreen ? SDL_FULLSCREEN : 0));

	if(modes == NULL)
		error("sdl: no modes available");

	_all_modes = (modes == (void*)-1);

	if (_all_modes) {
		// all modes available, put some default ones here
		memcpy(_resolutions, default_resolutions, sizeof(default_resolutions));
		_num_resolutions = lengthof(default_resolutions);
	} else {
		int n = 0;
		for(i = 0; modes[i]; i++) {
			int w = modes[i]->w;
			int h = modes[i]->h;
			if (IS_INT_INSIDE(w, 640, MAX_SCREEN_WIDTH + 1) &&
					IS_INT_INSIDE(h, 480, MAX_SCREEN_HEIGHT + 1)) {
				int j;
				for (j = 0; j < n; ++j)
					if (_resolutions[j][0] == w && _resolutions[j][1] == h)
						break;
				if (j == n) {
					_resolutions[n][0] = w;
					_resolutions[n][1] = h;
					if (++n == lengthof(_resolutions)) break;
				}
			}
		}
		_num_resolutions = n;
		qsort(_resolutions, n, sizeof(_resolutions[0]), compare_res);
	}
}

static int GetAvailableVideoMode(int *w, int *h)
{
	int i;
	int best;
	uint delta;

	// all modes available?
	if (_all_modes)
		return 1;

	// is the wanted mode among the available modes?
	for (i = 0; i != _num_resolutions; i++) {
		if(*w == _resolutions[i][0] && *h == _resolutions[i][1])
			return 1;
	}

	// use the closest possible resolution
	best = 0;
	delta = abs((_resolutions[0][0] - *w) * (_resolutions[0][1] - *h));
	for (i = 1; i != _num_resolutions; ++i) {
		uint newdelta = abs((_resolutions[i][0] - *w) * (_resolutions[i][1] - *h));
		if (newdelta < delta) {
			best = i;
			delta = newdelta;
		}
	}

	// use the default mode
	*w = _resolutions[best][0];
	*h = _resolutions[best][1];
	return 2;
}

static bool CreateMainSurface(int w, int h)
{
	SDL_Surface *newscreen;

	GetAvailableVideoMode(&w, &h);

	DEBUG(misc, 0) ("sdl: using mode %dx%d", w, h);


	// DO NOT CHANGE TO HWSURFACE, IT DOES NOT WORK + SDL_HWPALETTE
	newscreen = SDL_CALL SDL_SetVideoMode(w, h, 8, SDL_SWSURFACE  + (_fullscreen?SDL_FULLSCREEN:SDL_RESIZABLE));
	if(newscreen == NULL)
		return false;

	_screen.width = newscreen->w;
	_screen.height = newscreen->h;
	_screen.pitch = newscreen->pitch;

	_sdl_screen = newscreen;
	InitPalette();

	SDL_CALL SDL_WM_SetCaption("OpenTTD", "OpenTTD");
	SDL_CALL SDL_ShowCursor(0);

	GameSizeChanged();

	return true;
}

typedef struct {
	uint16 vk_from;
	byte vk_count;
	byte map_to;
} VkMapping;

#define AS(x,z) {x,0,z}
#define AM(x,y,z,w) {x,y-x,z}

static const VkMapping _vk_mapping[] = {
	// Pageup stuff + up/down
	AM(SDLK_PAGEUP,SDLK_PAGEDOWN, WKC_PAGEUP, WKC_PAGEDOWN),
	AS(SDLK_UP,			WKC_UP),
	AS(SDLK_DOWN,		WKC_DOWN),
	AS(SDLK_LEFT,		WKC_LEFT),
	AS(SDLK_RIGHT,	WKC_RIGHT),

	AS(SDLK_HOME,		WKC_HOME),
	AS(SDLK_END,		WKC_END),

	AS(SDLK_INSERT,	WKC_INSERT),
	AS(SDLK_DELETE,	WKC_DELETE),

	// Map letters & digits
	AM(SDLK_a,SDLK_z,'A','Z'),
	AM(SDLK_0,SDLK_9,'0','9'),

	AS(SDLK_ESCAPE,	WKC_ESC),
	AS(SDLK_BACKSPACE,	WKC_BACKSPACE),

	AS(SDLK_SPACE,		WKC_SPACE),
	AS(SDLK_RETURN,		WKC_RETURN),
	AS(SDLK_TAB,			WKC_TAB),

	// Function keys
	AM(SDLK_F1, SDLK_F12,	WKC_F1, WKC_F12),

	// Numeric part.
	// What is the virtual keycode for numeric enter??
	AM(SDLK_KP0,SDLK_KP9, WKC_NUM_0, WKC_NUM_9),
	AS(SDLK_KP_DIVIDE,		WKC_NUM_DIV),
	AS(SDLK_KP_MULTIPLY,	WKC_NUM_MUL),
	AS(SDLK_KP_MINUS,			WKC_NUM_MINUS),
	AS(SDLK_KP_PLUS,			WKC_NUM_PLUS),
	AS(SDLK_KP_ENTER,			WKC_NUM_ENTER),
	AS(SDLK_KP_PERIOD,		WKC_NUM_DECIMAL),
	{0, 0, 0}
};

static uint32 ConvertSdlKeyIntoMy(SDL_keysym *sym)
{
	const VkMapping	*map = _vk_mapping - 1;
	uint from;
	uint key = sym->sym;
	for(;;) {
		map++;
		from = map->vk_from;
		if (from == 0) {
			key = 0;
			break;
		}
		if ((uint)(key - from) <= map->vk_count) {
			key = key - from + map->map_to;
			break;
		}
	}

	// check scancode for BACKQUOTE key, because we want the key left of "1", not anything else (on non-US keyboards)
#if defined(WIN32)
	if (sym->scancode == 41) key |= WKC_BACKQUOTE;
#else
	#if defined(__APPLE__)
		if (sym->scancode == 10) key |= WKC_BACKQUOTE;
	#else
		#if defined(__MORPHOS__)
			if (sym->scancode == 0)  key |= WKC_BACKQUOTE;  // yes, that key is code '0' under MorphOS :)
		#else
			if (sym->scancode == 49) key |= WKC_BACKQUOTE;
		#endif
	#endif
#endif
	// META are the command keys on mac
	if (sym->mod & KMOD_META) key |= WKC_META;
	if (sym->mod & KMOD_SHIFT) key |= WKC_SHIFT;
	if (sym->mod & KMOD_CTRL) key |= WKC_CTRL;
	if (sym->mod & KMOD_ALT) key |= WKC_ALT;
	// these two lines really helps porting hotkey combos. Uncomment to use -- Bjarni
	DEBUG(misc,1) ("scancode character pressed %d\n", sym->scancode);
	DEBUG(misc,1) ("unicode character pressed %d\n", sym->unicode);
	return (key << 16) + sym->unicode;
}

static int jscount=0;

static int PollEvent() {
	SDL_Event ev;
	int xleft,yleft,xright,yright;
	if (!SDL_CALL SDL_PollEvent(&ev)) {
		int i;
		xleft=yleft=xright=yright=0;
		for(i=0;i<SDL_NumJoysticks();i++) {
            xleft=SDL_JoystickGetAxis(joy[i], 0);
			yleft=SDL_JoystickGetAxis(joy[i], 1);
			xright=SDL_JoystickGetAxis(joy[i], 2);
			yright=SDL_JoystickGetAxis(joy[i], 3);
			if(xright>100||xright<-100||yright>100||yright<-100||		
				xleft>100||xleft<-100||yleft>100||yleft<-100)
				break;
		}
		_cursor.joydelta.x=_cursor.joydelta.y=0;
		if(xright>100||xright<-100||yright>100||yright<-100) {
			if(xright>25000||xright<-25000||yright>25000||yright<-25000) {
				_cursor.joydelta.x=xright/500;
				_cursor.joydelta.y=yright/500;
			}
			else if(xright>20000||xright<-20000||yright>20000||yright<-20000) {
				_cursor.joydelta.x=xright/1000;
				_cursor.joydelta.y=yright/1000;
			}
			else if(xright>15000||xright<-15000||yright>15000||yright<-15000) {
				_cursor.joydelta.x=xright/1500;
				_cursor.joydelta.y=yright/1500;
			}
			else if(xright>10000||xright<-10000||yright>10000||yright<-10000) {
				_cursor.joydelta.x=xright/2000;
				_cursor.joydelta.y=yright/2000;
			}
			else if(xright>5000||xright<-5000||yright>5000||yright<-5000) {
				_cursor.joydelta.x=xright/3000;
				_cursor.joydelta.y=yright/3000;
			}
			else {
				_cursor.joydelta.x=xright/4000;
				_cursor.joydelta.y=yright/4000;
			}
			_scrolling_viewport=true;
		}
#if 1
		if((xleft>100||xleft<-100||yleft>100||yleft<-100)&&++jscount>=3) {
			int dx=0,dy=0;
			jscount=0;
			if(xleft>22000||xleft<-22000||yleft>22000||yleft<-22000) {
				xleft=xleft > 22000 ? 22000 : xleft < -22000 ? -22000 : xleft;
				yleft=yleft > 22000 ? 22000 : yleft < -22000 ? -22000 : yleft;
			}
			if(xleft>20000||xleft<-20000||yleft>20000||yleft<-20000) {
				dx=xleft/1500;
				dy=yleft/1500;
			}
			else if(xleft>15000||xleft<-15000||yleft>15000||yleft<-15000) {
				dx=xleft/3000;
				dy=yleft/3000;
			}
			else {
				if(xleft>10000)
					dx=2;
				else if (xleft<-10000)
					dx=-2;
				if(yleft>10000)
					dy=2;
				else if(yleft<-10000)
					dy=-2;
				if(dy!=0||dx!=0)
					jscount=-3;
			}
			ev.motion.type=SDL_MOUSEMOTION;
			ev.motion.x=_cursor.pos.x+dx;
			ev.motion.y=_cursor.pos.y+dy;
			if((_cursor.pos.x+dx)<2)
				ev.motion.x=1;
			if(ev.motion.x>_screen.width-2)
				ev.motion.x=_screen.width-1;
			if((_cursor.pos.y+dy)<2)
				ev.motion.y=1;
			if(ev.motion.y>_screen.height-2)
				ev.motion.y=_screen.height-1;
			ev.motion.state=0;
			ev.motion.which=5;
			ev.motion.xrel=dx;
			ev.motion.yrel=dy;
		}
#endif
#if 0
		if((xleft>100||xleft<-100||yleft>100||yleft<-100)&&++jscount>=3) {
			else if( (xleft>10000&&xleft<17000) || 
					 (xleft<-10000&&xleft>-17000) || 
					 (yleft>10000&&yleft<17000) || 
					 (yleft<-10000&&yleft>-17000))
				dx=xleft-10000;
				dy=yleft3000;
			}
		}
#endif
		else
			return -2;
	}

	switch(ev.type) {
	case SDL_JOYBUTTONDOWN:
		DEBUG(misc,1) ("Joystick button pressed: id: %d state: %d",ev.jbutton.button,ev.jbutton.state);
		switch(ev.jbutton.button) {
			case 6:
			case 0:
				_left_button_down = true;
				break;
			case 7:
			case 1:
				_right_button_down = true;
				_right_button_clicked = true;
				break;
			case 3: // Y
				PlaceLandBlockInfo();
				_left_button_down = true;
				_info_button_down=true;
				break;
			case 2: // X
			case 4: // white
			case 5: // black
				break;
		}
		break;
	case SDL_JOYBUTTONUP:
		switch(ev.jbutton.button) {
			case 6:
			case 0:
				_left_button_down = false;
				_left_button_clicked = false;
				break;
			case 7:
			case 1:
				_right_button_down = false;
				break;
			case 3: // Y infoclick
				ResetObjectToPlace();
				DeleteWindowById(WC_LAND_INFO, 0);
				_left_button_down = false;
				_left_button_clicked = false;
				break;
			case 2: // X delete click ?
			case 4: // white terrain up click? 
			case 5: // black terrain down click?
				break;
		}
		break;
	case SDL_MOUSEMOTION:
		if (_cursor.fix_at) {
			int dx = ev.motion.x - _cursor.pos.x; // ev.motion.xrel;
			int dy = ev.motion.y - _cursor.pos.y; // ev.motion.yrel;
			if (dx != 0 || dy != 0) {
				_cursor.delta.x += dx;
				_cursor.delta.y += dy;
				SDL_CALL SDL_WarpMouse(_cursor.pos.x, _cursor.pos.y);
			}
		} else {
			_cursor.delta.x = ev.motion.xrel;
			_cursor.delta.y = ev.motion.yrel;
			_cursor.pos.x += _cursor.delta.x;
			_cursor.pos.y += _cursor.delta.y;
			if(_cursor.pos.x<2)
				_cursor.pos.x=1;
			if(_cursor.pos.x>_screen.width-2)
				_cursor.pos.x=_screen.width-1;
			if(_cursor.pos.y<2)
				_cursor.pos.y=1;
			if(_cursor.pos.y>_screen.height-2)
				_cursor.pos.y=_screen.height-1;
			_cursor.dirty = true;
		}
		break;

	case SDL_MOUSEBUTTONDOWN:
		if (_rightclick_emulate && (SDL_CALL SDL_GetModState() & (KMOD_LCTRL | KMOD_RCTRL)))
			ev.button.button = SDL_BUTTON_RIGHT;

		if (ev.button.button == SDL_BUTTON_LEFT) {
			_left_button_down = true;
		} else if (ev.button.button == SDL_BUTTON_RIGHT) {
			_right_button_down = true;
			_right_button_clicked = true;
		}
#if !defined(WIN32)
		 else if (ev.button.button == SDL_BUTTON_WHEELUP) {
			_cursor.wheel--;
		} else if (ev.button.button == SDL_BUTTON_WHEELDOWN) {
			_cursor.wheel++;
		}
#endif
		break;

	case SDL_MOUSEBUTTONUP:
		if (_rightclick_emulate) {
			_right_button_down = false;
			_left_button_down = false;
			_left_button_clicked = false;
		} else if (ev.button.button == SDL_BUTTON_LEFT) {
			_left_button_down = false;
			_left_button_clicked = false;
		} else if (ev.button.button == SDL_BUTTON_RIGHT) {
			_right_button_down = false;
		}
		break;

	case SDL_QUIT:
		// do not ask to quit on the main screen
		if(_game_mode != GM_MENU)
			AskExitGame();
		else
			return ML_QUIT;
		break;

	case SDL_KEYDOWN:
		if ((((ev.key.keysym.sym == SDLK_RETURN) || (ev.key.keysym.sym == SDLK_f)) && (ev.key.keysym.mod & KMOD_ALT)) || (((ev.key.keysym.sym == SDLK_RETURN) || (ev.key.keysym.sym == SDLK_f)) && (ev.key.keysym.mod & KMOD_META))) {
			_fullscreen ^= true;
			GetVideoModes();
			CreateMainSurface(_screen.width, _screen.height);
			MarkWholeScreenDirty();
		} else {
			_pressed_key = ConvertSdlKeyIntoMy(&ev.key.keysym);
		}
		break;

	case SDL_VIDEORESIZE: {
			int w, h;
			w = ev.resize.w;
			h = ev.resize.h;

			w = clamp(w, 64, MAX_SCREEN_WIDTH);
			h = clamp(h, 64, MAX_SCREEN_HEIGHT);

			ChangeResInGame(w, h);

			break;
		}
	}
	return -1;
}

extern float xoffset,yoffset,xstretch,ystretch;

static const char *SdlVideoStart(char **parm)
{
    char buf[30];
	int i;

	{char *s;if ((s = SdlOpen(SDL_INIT_VIDEO)) != NULL) return s;}

	SDL_CALL SDL_VideoDriverName(buf, 30);
	DEBUG(misc, 0) ("sdl: using driver '%s'", buf);

	GetVideoModes();
	CreateMainSurface(_cur_resolution[0], _cur_resolution[1]);
	MarkWholeScreenDirty();
    SDL_XBOX_SetScreenPosition(xoffset,yoffset);
    SDL_XBOX_SetScreenStretch(xstretch,ystretch);
	{char *s;if ((s = SdlOpen(SDL_INIT_JOYSTICK)) != NULL) return s;}
	SDL_CALL SDL_JoystickEventState(SDL_ENABLE);
	DEBUG(misc,1) ("Num joysticks: %d",SDL_NumJoysticks());
	for (i=0;i<SDL_NumJoysticks();i++) {
		DEBUG(misc,1) ("Opening joystick %d");
		joy[i]=SDL_JoystickOpen(i);
		DEBUG(misc,1) ("Name: %s\n", SDL_CALL SDL_JoystickName(i));
	    DEBUG(misc,1) ("Number of Axes: %d\n", SDL_CALL SDL_JoystickNumAxes(joy[i]));
	    DEBUG(misc,1) ("Number of Buttons: %d\n", SDL_CALL SDL_JoystickNumButtons(joy[i]));
	    DEBUG(misc,1) ("Number of Balls: %d\n", SDL_CALL SDL_JoystickNumBalls(joy[i]));
		DEBUG(misc,1) ("Number of hats: %d\n", SDL_CALL SDL_JoystickNumHats(joy[i]));
	}
	SDL_CALL SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_CALL SDL_EnableUNICODE(1);
	return NULL;
}

static void SdlVideoStop()
{
	SdlClose(SDL_INIT_VIDEO);
}

static int SdlVideoMainLoop()
{
	uint32 next_tick = SDL_CALL SDL_GetTicks() + 30, cur_ticks, pal_tick = 0;
	int i;
	uint32 mod;
	int numkeys;
	Uint8 *keys;

	while (true) {
		InteractiveRandom(); // randomness

		while ((i=PollEvent()) == -1) {}
		if (i>=0) return i;

		if (_exit_game)	return ML_QUIT;

		mod = SDL_CALL SDL_GetModState();
		keys = SDL_CALL SDL_GetKeyState(&numkeys);
#if defined(_DEBUG)
		if (_shift_pressed) {
#else
		if (keys[SDLK_TAB]) {
#endif
			if (!_networking) _fast_forward |= 2;
		} else if (_fast_forward&2) {
			_fast_forward = 0;
		}

		cur_ticks=SDL_CALL SDL_GetTicks();
		if ((_fast_forward && !_pause) || cur_ticks > next_tick)
			next_tick = cur_ticks;

		if (cur_ticks == next_tick) {
			next_tick += 30;

			_ctrl_pressed = !!(mod & (KMOD_LCTRL | KMOD_RCTRL));
			_shift_pressed = !!(mod & (KMOD_LSHIFT | KMOD_RSHIFT));
			_dbg_screen_rect = !!(mod & KMOD_CAPS);

			// determine which directional keys are down
			_dirkeys =
			    (keys[SDLK_LEFT] ? 1 : 0) +
			    (keys[SDLK_UP] ? 2 : 0) +
			    (keys[SDLK_RIGHT] ? 4 : 0) +
			    (keys[SDLK_DOWN] ? 8 : 0);
			GameLoop();
			DEBUG(misc,8) ("Did an SDL gameloop");
			_screen.dst_ptr = _sdl_screen->pixels;
			UpdateWindows();
			DEBUG(misc,8) ("(SDL) Updated windows");
			if (++ pal_tick > 4){
			    CheckPaletteAnim();
			    pal_tick = 1;
				DEBUG(misc,8) ("(SDL) Checked palette");
			}
 			DrawSurfaceToScreen();
			DEBUG(misc,8) ("Did an SDL tick");
		} else {
			SDL_CALL SDL_Delay(1);
			_screen.dst_ptr = _sdl_screen->pixels;
			DrawMouseCursor();
			DrawSurfaceToScreen();
			DEBUG(misc,8) ("Did an SDL delay");
		}
	}
}

static bool SdlVideoChangeRes(int w, int h)
{
	// see if the mode is available
	if (GetAvailableVideoMode(&w, &h) != 1)
		return false;

	CreateMainSurface(w, h);
	return true;
}

const HalVideoDriver _sdl_video_driver = {
	SdlVideoStart,
	SdlVideoStop,
	SdlVideoMakeDirty,
	SdlVideoMainLoop,
	SdlVideoChangeRes,
};

// MRFMRF SOUND START
SDL_AudioSpec spec;

static char *SDLAudioStart() {
	{char *s;if ((s = SdlOpen(SDL_INIT_AUDIO)) != NULL) return s;}
}

static void CDECL fill_sound_buffer(void *userdata, Uint8 *stream, int len)
{
	MxMixSamples(_mixer, stream, len >> 2);
}

static char *SdlSoundStart(char **parm)
{

	spec.freq = GetDriverParamInt(parm, "hz", 11025);
	spec.format = AUDIO_S16SYS;
	spec.channels = 2;
	spec.samples = 512;
	//*(void**)&spec.callback = fill_sound_buffer;
// commented -> music driver will open audio.
//	SDL_CALL SDL_OpenAudio(&spec, &spec); 
//	SDL_CALL SDL_PauseAudio(0);
	return NULL;
}

static void SdlSoundStop()
{
	SDL_CALL SDL_CloseAudio();
	SdlClose(SDL_INIT_AUDIO);
}

const HalSoundDriver _sdl_sound_driver = {
	SdlSoundStart,
	SdlSoundStop,
};

Mix_Music *music = NULL;
struct music_data {
	char *filename;
	Mix_Music *music;
} mdata[30];

static char *sdl_midi_start(char **parm) {
  int audio_rate = 44100;//= GetDriverParamInt(parm, "hz", 11025);
  Uint16 audio_format = AUDIO_S16SYS; /* 16-bit stereo */
  int audio_channels = 2;
  int audio_buffers = 4096;
  int i;
  SDL_Init(SDL_INIT_AUDIO);
  if(Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers)) {
    printf("Unable to open midi!\n");
  }
  Mix_SetPostMix(fill_sound_buffer,0);
  memset(mdata,0,sizeof(mdata));
  // Precache/load all midis on startup, set to 1 to enable (warning, too slow.)
#if 0
  for(i=0;i<22;i++) { // standard midi ranges from 0-21
	  char buf[200];
	  sprintf(buf, "%sgm_tt%.2d.mid",  _path.gm_dir, i);
	  mdata[i].filename=strdup(buf);
	  mdata[i].music=Mix_LoadMUS(mdata[i].filename);
  }
#endif
	
  return 0;
}


static void sdl_midi_stop(void) {
	if(music!=NULL)	{
		Mix_HaltMusic();
//		Mix_FreeMusic(music);
		music = NULL;
    }
	Mix_CloseAudio();
}

static void sdl_midi_song_done(void) {
	if(music!=NULL)	{
		Mix_HaltMusic();
//		Mix_FreeMusic(music);
		music = NULL;
    }
}

static void sdl_midi_play_song(const char *filename) {
	int i;
	if(music!=NULL)	{
		Mix_HaltMusic();
//		Mix_FreeMusic(music);
		music = NULL;
    } // loading music takes a lot of time so cache them.
	// its still annoying when loading them for the first time, maybe load all on start.
	for(i=0;i<30;i++) {
		if(!mdata[i].filename) { // end of line
			mdata[i].filename=strdup(filename);
			mdata[i].music=Mix_LoadMUS(filename);
			break;
		}
		else if(!strcmp(mdata[i].filename,filename)) { // playing something we cached.
			break;
		}
	}
	music = mdata[i].music; // Mix_LoadMUS(filename);
	Mix_PlayMusic(music, 0);
	Mix_HookMusicFinished(sdl_midi_song_done);
}

static void sdl_midi_stop_song(void) {
	if(music!=NULL)	{
		Mix_HaltMusic();
//		Mix_FreeMusic(music);
		music = NULL;
    }
}

static bool sdl_midi_is_playing(void) {
	return music!=NULL;
}

static void sdl_midi_set_volume(byte vol) {
	Mix_VolumeMusic(vol) ;
}

const HalMusicDriver _sdl_midi_music_driver = {
	sdl_midi_start,
	sdl_midi_stop,
	sdl_midi_play_song,
	sdl_midi_stop_song,
	sdl_midi_is_playing,
	sdl_midi_set_volume,
};


#include "viewport.h"
void redsq_debug(int tile)
{
	_thd.redsq = tile;
	MarkWholeScreenDirty();
	_screen.dst_ptr = _sdl_screen->pixels;
	UpdateWindows();

	SdlVideoMakeDirty(0,0,_screen.width,_screen.height);
	DrawSurfaceToScreen();
}

void DbgRedraw()
{
	SdlVideoMakeDirty(0,0,_screen.width,_screen.height);
	DrawSurfaceToScreen();
}
#endif // WITH_SDL
